import asyncio
import functools
from functools import wraps
from typing import Awaitable, Callable, Optional, Type

from langgraph.func import task
from langgraph.pregel.call import SyncAsyncFuture
from langgraph.types import RetryPolicy

from livechain.graph.cron import CronExpr
from livechain.graph.func.routine import (
    CronSignalRoutine,
    EventSignalRoutine,
    Mode,
    ReactiveSignalRoutine,
    SignalRoutineRunner,
    SignalStrategy,
    WorkflowSignalRoutine,
)
from livechain.graph.types import (
    CronEffect,
    CronSignal,
    EntrypointFunc,
    P,
    ReactiveEffect,
    ReactiveSignal,
    Subscriber,
    T,
    TEvent,
    TriggerSignal,
    TState,
    WatchedValue,
)


def step(
    *,
    name: Optional[str] = None,
    retry: Optional[RetryPolicy] = None,
):
    def step_wrapper(
        func: Callable[P, Awaitable[T]],
    ) -> Callable[P, SyncAsyncFuture[T]]:
        func_name = name if name is not None else func.name if isinstance(func, SignalRoutineRunner) else func.__name__

        if not asyncio.iscoroutinefunction(func) and not isinstance(func, SignalRoutineRunner):
            raise ValueError("Step function must be async or a SignalRoutineRunner, got %s" % type(func))

        @task(name=func_name, retry=retry)
        async def step_wrapper_task(*args: P.args, **kwargs: P.kwargs) -> T:
            result = await func(*args, **kwargs)  # type: ignore
            return result

        task_func = functools.update_wrapper(step_wrapper_task, func)
        return task_func  # type: ignore

    return step_wrapper


def subscribe(
    event_schema: Type[TEvent],
    *,
    strategy: Optional[SignalStrategy] = None,
    name: Optional[str] = None,
    retry: Optional[RetryPolicy] = None,
) -> Callable[[Subscriber[TEvent]], EventSignalRoutine[TEvent]]:
    def subscribe_decorator(
        subscriber: Subscriber[TEvent],
    ) -> EventSignalRoutine[TEvent]:
        return EventSignalRoutine(
            schema=event_schema,
            routine=subscriber,
            strategy=strategy,
            name=name,
            retry=retry,
        )

    return subscribe_decorator


def reactive(
    state_schema: Type[TState],
    cond: WatchedValue[TState, T],
    *,
    strategy: Optional[SignalStrategy] = None,
    name: Optional[str] = None,
    retry: Optional[RetryPolicy] = None,
) -> Callable[[ReactiveEffect[TState]], ReactiveSignalRoutine[TState, T]]:
    def reactive_decorator(
        effect: ReactiveEffect[TState],
    ) -> ReactiveSignalRoutine[TState, T]:
        @wraps(effect)
        async def effect_wrapper(signal: ReactiveSignal[TState]):
            return await effect(signal.old_state, signal.new_state)

        return ReactiveSignalRoutine(
            schema=ReactiveSignal[state_schema],
            routine=effect_wrapper,
            state_schema=state_schema,
            cond=cond,
            strategy=strategy,
            name=name,
            retry=retry,
        )

    return reactive_decorator


def cron(
    expr: CronExpr,
    *,
    strategy: Optional[SignalStrategy] = None,
    name: Optional[str] = None,
    retry: Optional[RetryPolicy] = None,
) -> Callable[[CronEffect], CronSignalRoutine]:
    def cron_decorator(
        cron_effect: CronEffect,
    ) -> CronSignalRoutine:
        @wraps(cron_effect)
        async def cron_wrapper(signal: CronSignal):
            return await cron_effect()

        return CronSignalRoutine(
            schema=CronSignal,
            cron_expr=expr,
            routine=cron_wrapper,
            strategy=strategy,
            name=name,
            retry=retry,
        )

    return cron_decorator


def root(*, name: Optional[str] = None, retry: Optional[RetryPolicy] = None):
    def root_decorator(func: EntrypointFunc) -> WorkflowSignalRoutine:
        @wraps(func)
        async def workflow_wrapper(trigger: TriggerSignal):
            return await func()

        return WorkflowSignalRoutine(
            schema=TriggerSignal,
            routine=workflow_wrapper,
            strategy=Mode.Interrupt(),
            name=name,
            retry=retry,
        )

    return root_decorator



from __future__ import annotations

import asyncio
import logging
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Any, Awaitable, Callable, Dict, Generic, Optional, Type, Union

from langchain_core.runnables import Runnable, RunnableConfig
from langgraph.func import entrypoint
from langgraph.types import RetryPolicy
from pydantic import TypeAdapter, ValidationError

from livechain.graph.constants import SENTINEL
from livechain.graph.cron import CronExpr
from livechain.graph.types import (
    CronSignal,
    LangGraphInjectable,
    ReactiveSignal,
    T,
    TEvent,
    TModel,
    TriggerSignal,
    TState,
    WatchedValue,
)

logger = logging.getLogger(__name__)


class Mode:
    @dataclass
    class Interrupt:
        pass

    @dataclass
    class Parallel:
        pass

    @dataclass
    class Queue:
        pass

    @dataclass
    class Debounce:
        delay: float


SignalStrategy = Union[
    Mode.Interrupt,
    Mode.Parallel,
    Mode.Queue,
    Mode.Debounce,
]


class SignalRoutineType(str, Enum):
    SUBSCRIBE = "EventCallback"
    REACTIVE = "ReactiveEffect"
    CRON = "CronEffect"
    WORKFLOW = "Workflow"


def default_signal_strategy() -> Mode.Parallel:
    return Mode.Parallel()


class BaseSignalRoutine(Generic[TModel], ABC):
    def __init__(
        self,
        schema: Type[TModel],
        routine: Callable[[TModel], Awaitable[None]],
        strategy: Optional[SignalStrategy] = None,
        name: Optional[str] = None,
        retry: Optional[RetryPolicy] = None,
    ):
        self._schema = schema
        self._routine = routine
        self._strategy = strategy or default_signal_strategy()
        self._name = name if name is not None else self._routine.__name__
        self._retry = retry

    @property
    @abstractmethod
    def routine_type(self) -> SignalRoutineType:
        raise NotImplementedError

    @property
    def schema(self) -> Type[TModel]:
        return self._schema

    @property
    def name(self) -> str:
        return self._name

    @property
    def mode(self) -> SignalStrategy:
        return self._strategy

    def create_routine_runnable(
        self,
        injectable: LangGraphInjectable | None = None,
    ) -> Runnable[TModel, Any]:
        from livechain.graph.func import step
        from livechain.graph.func.utils import rename_function

        injectable = injectable or LangGraphInjectable.from_empty()

        @step(name=self._name, retry=self._retry)
        async def routine_step(signal: TModel):
            return await self._routine(signal)

        @entrypoint(
            checkpointer=injectable.checkpointer,
            store=injectable.store,
            config_schema=injectable.config_schema,
        )
        @rename_function(self.routine_type.value)
        async def routine_entrypoint(signal: TModel):
            return await routine_step(signal)

        return routine_entrypoint

    def create_runner(
        self,
        config: RunnableConfig | None = None,
        injectable: LangGraphInjectable | None = None,
    ) -> SignalRoutineRunner[TModel]:
        injectable = injectable or LangGraphInjectable.from_empty()
        routine_runnable = self.create_routine_runnable(injectable)

        runner_cls: Optional[Type[SignalRoutineRunner[TModel]]] = {
            Mode.Interrupt: InterruptableSignalRoutineRunner,
            Mode.Parallel: ParallelSignalRoutineRunner,
            Mode.Queue: FifoSignalRoutineRunner,
            Mode.Debounce: DebounceSignalRoutineRunner,
        }.get(type(self._strategy))

        if runner_cls is None:
            raise ValueError(f"Invalid signal routine strategy: {self._strategy}")

        return runner_cls(
            self._schema,
            routine_runnable,
            self._strategy,
            config,
            self._name,
        )


class WorkflowSignalRoutine(BaseSignalRoutine[TriggerSignal]):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def routine_type(self) -> SignalRoutineType:
        return SignalRoutineType.WORKFLOW


class EventSignalRoutine(BaseSignalRoutine[TEvent]):
    @property
    def routine_type(self) -> SignalRoutineType:
        return SignalRoutineType.SUBSCRIBE


class ReactiveSignalRoutine(BaseSignalRoutine[ReactiveSignal[TState]], Generic[TState, T]):
    def __init__(
        self,
        schema: Type[ReactiveSignal[TState]],
        routine: Callable[[ReactiveSignal[TState]], Awaitable[None]],
        state_schema: Type[TState],
        cond: WatchedValue[TState, T],
        name: Optional[str] = None,
        strategy: Optional[SignalStrategy] = None,
        retry: Optional[RetryPolicy] = None,
    ):
        super().__init__(schema, routine, strategy, name, retry)
        self._state_schema = state_schema
        self._cond = cond

    @property
    def routine_type(self) -> SignalRoutineType:
        return SignalRoutineType.REACTIVE

    @property
    def cond(self) -> WatchedValue[TState, T]:
        return self._cond

    @property
    def state_schema(self) -> Type[TState]:
        return self._state_schema


class CronSignalRoutine(BaseSignalRoutine[CronSignal]):
    def __init__(
        self,
        schema: Type[CronSignal],
        routine: Callable[[CronSignal], Awaitable[Any]],
        cron_expr: CronExpr,
        strategy: Optional[SignalStrategy] = None,
        name: Optional[str] = None,
        retry: Optional[RetryPolicy] = None,
    ):
        super().__init__(schema, routine, strategy, name, retry)
        self._cron_expr = cron_expr

    @property
    def routine_type(self) -> SignalRoutineType:
        return SignalRoutineType.CRON

    @property
    def cron_expr(self) -> CronExpr:
        return self._cron_expr


class SignalRoutineRunner(Generic[TModel], ABC):
    def __init__(
        self,
        schema: Type[TModel],
        runnable: Runnable[TModel, None],
        strategy: SignalStrategy,
        config: RunnableConfig,
        name: str,
    ):
        self._id = uuid.uuid4()
        self._schema = schema
        self._runnable = runnable
        self._strategy = strategy
        self._config = config
        self._name = name
        self._signal_queue = asyncio.Queue()

    @property
    def schema(self) -> Type[TModel]:
        return self._schema

    @property
    def name(self) -> str:
        return self._name

    @property
    def routine_id(self) -> str:
        return str(self._id)

    @property
    def strategy(self) -> SignalStrategy:
        return self._strategy

    async def __call__(self, signal: TModel):
        try:
            adapter = TypeAdapter(self._schema)
            validated_signal = adapter.validate_python(signal)
            await self._signal_queue.put(validated_signal)
        except ValidationError as e:
            logger.error(f"Routine runner {self._name} of id {self.routine_id} received invalid data: {e}")

    @abstractmethod
    async def start(self):
        raise NotImplementedError

    @abstractmethod
    def stop(self):
        raise NotImplementedError


class InterruptableSignalRoutineRunner(SignalRoutineRunner[TModel]):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._current_task: Optional[asyncio.Task] = None

    async def _start_routine_with_interrupts(self):
        while True:
            signal = await self._signal_queue.get()

            if signal is SENTINEL:
                break

            try_cancel_asyncio_task(self._current_task)
            self._current_task = asyncio.create_task(self._runnable.ainvoke(signal, config=self._config))

        try_cancel_asyncio_task(self._current_task)
        logger.info(f"Routine runner {self._name} of id {self.routine_id} stopped")

    async def start(self):
        await self._start_routine_with_interrupts()

    def stop(self):
        self._signal_queue.put_nowait(SENTINEL)


class ParallelSignalRoutineRunner(SignalRoutineRunner[TModel]):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._tasks: Dict[uuid.UUID, asyncio.Task] = {}

    def _on_task_done(self, task_id: uuid.UUID):
        self._tasks.pop(task_id)

    def _cancel_tasks(self):
        for task in self._tasks.values():
            try_cancel_asyncio_task(task)

    async def _start_routine_in_parallel(self):
        while True:
            signal = await self._signal_queue.get()

            if signal is SENTINEL:
                break

            task_id = uuid.uuid4()
            task = asyncio.create_task(self._runnable.ainvoke(signal, config=self._config))
            task.add_done_callback(lambda _, tid=task_id: self._on_task_done(tid))
            self._tasks[task_id] = task

        self._cancel_tasks()
        logger.info(f"Routine runner {self._name} of id {self.routine_id} stopped")

    async def start(self):
        await self._start_routine_in_parallel()

    def stop(self):
        self._signal_queue.put_nowait(SENTINEL)
        self._cancel_tasks()


class FifoSignalRoutineRunner(SignalRoutineRunner[TModel]):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._current_task: Optional[asyncio.Task] = None

    async def _start_routine_in_fifo(self):
        while True:
            signal = await self._signal_queue.get()

            if signal is SENTINEL:
                break

            try:
                self._current_task = asyncio.create_task(self._runnable.ainvoke(signal, config=self._config))
                await self._current_task
            except asyncio.CancelledError:
                pass
            except Exception as e:
                logger.error(f"Routine runner {self._name} of id {self.routine_id} received an exception: {e}")

        try_cancel_asyncio_task(self._current_task)
        logger.info(f"Routine runner {self._name} of id {self.routine_id} stopped")

    async def start(self):
        await self._start_routine_in_fifo()

    def stop(self):
        self._signal_queue.put_nowait(SENTINEL)
        try_cancel_asyncio_task(self._current_task)


class DebounceSignalRoutineRunner(SignalRoutineRunner[TModel]):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._current_task: Optional[asyncio.Task] = None
        self._delay = self._strategy.delay  # type: ignore
        self._counter: int = 0

    async def _process_with_delay(self, signal: TModel, counter: int):
        await asyncio.sleep(self._delay)
        if counter == self._counter:
            await self._runnable.ainvoke(signal, config=self._config)

    async def start(self):
        while True:
            signal = await self._signal_queue.get()
            self._counter += 1

            if signal is SENTINEL:
                break

            try_cancel_asyncio_task(self._current_task)
            self._current_task = asyncio.create_task(self._process_with_delay(signal, self._counter))

        try_cancel_asyncio_task(self._current_task)
        logger.info(f"Routine runner {self._name} of id {self.routine_id} stopped")

    def stop(self):
        self._signal_queue.put_nowait(SENTINEL)
        try_cancel_asyncio_task(self._current_task)


def try_cancel_asyncio_task(task: Optional[asyncio.Task]):
    if task is not None and not task.done():
        task.cancel()



from livechain.graph.emitter import Emitter
from livechain.graph.executor import Workflow, WorkflowExecutor
from livechain.graph.func import cron, reactive, root, step, subscribe
from livechain.graph.ops import channel_send, mutate_state, publish_event, trigger_workflow
from livechain.graph.types import EventSignal, ReactiveSignal, TriggerSignal

__all__ = [
    "cron",
    "reactive",
    "root",
    "step",
    "subscribe",
    "channel_send",
    "mutate_state",
    "publish_event",
    "trigger_workflow",
    "EventSignal",
    "ReactiveSignal",
    "TriggerSignal",
    "Emitter",
    "Workflow",
    "WorkflowExecutor",
]



from typing import Any, Dict, Type

from langgraph.checkpoint.memory import MemorySaver
from langgraph.graph import START, StateGraph
from langgraph.graph.state import CompiledStateGraph
from pydantic import PrivateAttr

from livechain.graph.persist.base import BaseStatePersister
from livechain.graph.types import TState
from livechain.graph.utils import make_config

CONFIG = make_config({"thread_id": "1"})


def create_base_graph(state_schema: Type[TState]):
    def save_state(_: Any):
        return None

    graph = (
        StateGraph(state_schema=state_schema)
        .add_node("save_state", save_state)
        .add_edge(START, "save_state")
        .compile(checkpointer=MemorySaver())
    )

    return graph


class LocalStatePersister(BaseStatePersister[TState]):
    _graph: CompiledStateGraph = PrivateAttr()

    def __init__(self, state_schema: Type[TState]):
        super().__init__(state_schema=state_schema)
        self._graph = create_base_graph(state_schema)

    def _get(self) -> TState:
        return self.state_schema.model_validate(self._graph.get_state(CONFIG).values)

    def _set(self, state: TState | Dict[str, Any]) -> TState:
        if isinstance(state, dict):
            state_patch = state
        else:
            state_patch = self.state_schema.model_dump(state)

        raw_state = self._graph.invoke(state_patch, CONFIG)
        return self.state_schema.model_validate(raw_state)



import asyncio
from typing import Any, Awaitable, Callable, Dict, Generic, Optional, Set

from pydantic import BaseModel, ConfigDict, PrivateAttr

from livechain.graph.func.utils import async_parallel
from livechain.graph.types import T, THashable


class Emitter(BaseModel, Generic[THashable, T]):
    get_hash: Callable[[T], THashable]

    _subscribers: Dict[THashable, Set[Callable[[T], Awaitable[Any]]]] = PrivateAttr(default_factory=dict)

    _default_subscribers: Set[Callable[[T], Awaitable[Any]]] = PrivateAttr(default_factory=set)

    _callback_to_hash: Dict[Callable[[T], Awaitable[Any]], Optional[THashable]] = PrivateAttr(default_factory=dict)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def subscribe(
        self,
        data: Optional[THashable] = None,
        *,
        callback: Callable[[T], Awaitable[Any]],
    ) -> None:
        if data is not None:
            self._subscribers.setdefault(data, set()).add(callback)
        else:
            self._default_subscribers.add(callback)
        self._callback_to_hash[callback] = data

    def emit(self, data: T) -> asyncio.Task[Dict[str, Any]]:
        data_hash = self.get_hash(data)
        callbacks = self._subscribers.get(data_hash, [])

        func_map: Dict[str, Callable[[T], Awaitable[Any]]] = {}
        for i, callback in enumerate(callbacks):
            func_map[f"callback_{i}"] = callback

        for i, callback in enumerate(self._default_subscribers):
            func_map[f"default_callback_{i}"] = callback

        runnable = async_parallel(func_map)
        return asyncio.create_task(runnable.ainvoke(data))

    def unsubscribe(self, callback: Callable[[T], Awaitable[Any]]) -> None:
        if callback not in self._callback_to_hash:
            return

        data = self._callback_to_hash[callback]
        if data is None:
            self._default_subscribers.remove(callback)
        else:
            self._subscribers[data].remove(callback)

    def unsubscribe_all(self) -> None:
        self._subscribers.clear()
        self._default_subscribers.clear()
        self._callback_to_hash.clear()


def emitter_factory(
    get_hash: Callable[[T], THashable],
) -> Callable[[], Emitter[THashable, T]]:
    def create_emitter() -> Emitter[THashable, T]:
        return Emitter(get_hash=get_hash)

    return create_emitter



import asyncio
from unittest.mock import AsyncMock

import pytest

from livechain.graph.emitter import emitter_factory
from livechain.graph.types import EventSignal


class IntEvent(EventSignal):
    data: int


class StrEvent(EventSignal):
    data: str


@pytest.fixture
def event_emitter():
    def get_event_hash(event: IntEvent):
        return type(event)

    return emitter_factory(get_event_hash)()


@pytest.mark.asyncio
async def test_emitter_register_default_callback(event_emitter):
    # Create an async mock function
    mock_callback = AsyncMock()

    # Subscribe the mock to the emitter
    event_emitter.subscribe(callback=mock_callback)
    event = IntEvent(data=42)

    # Emit an event
    await event_emitter.emit(event)

    # Check if the mock was called
    mock_callback.assert_called_once_with(event)


@pytest.mark.asyncio
async def test_emitter_register_default_callbacks(event_emitter):
    mock_callbacks = [AsyncMock() for _ in range(3)]

    for callback in mock_callbacks:
        event_emitter.subscribe(callback=callback)

    event = IntEvent(data=42)

    await event_emitter.emit(event)

    for callback in mock_callbacks:
        callback.assert_called_once_with(event)


@pytest.mark.asyncio
async def test_emitter_register_callbacks(event_emitter):
    mock_callbacks = [AsyncMock() for _ in range(3)]

    for callback in mock_callbacks:
        event_emitter.subscribe(IntEvent, callback=callback)

    event = IntEvent(data=42)

    await event_emitter.emit(event)

    for callback in mock_callbacks:
        callback.assert_called_once_with(event)


@pytest.mark.asyncio
async def test_emitter_trigger_specific_callback(event_emitter):
    mock_callback_int = AsyncMock()
    mock_callback_str = AsyncMock()

    event_emitter.subscribe(IntEvent, callback=mock_callback_int)
    event_emitter.subscribe(StrEvent, callback=mock_callback_str)

    int_event = IntEvent(data=42)

    await event_emitter.emit(int_event)

    mock_callback_int.assert_called_once_with(int_event)
    mock_callback_str.assert_not_called()

    str_event = StrEvent(data="test")

    await event_emitter.emit(str_event)

    mock_callback_int.assert_called_once_with(int_event)
    mock_callback_str.assert_called_once_with(str_event)


@pytest.mark.asyncio
async def test_emitter_trigger_both_default_and_specific_callbacks(event_emitter):
    mock_callback_default_1 = AsyncMock()
    mock_callback_default_2 = AsyncMock()
    mock_callback_int = AsyncMock()

    event_emitter.subscribe(callback=mock_callback_default_1)
    event_emitter.subscribe(callback=mock_callback_default_2)
    event_emitter.subscribe(IntEvent, callback=mock_callback_int)

    event = IntEvent(data=42)

    await event_emitter.emit(event)

    mock_callback_default_1.assert_called_once_with(event)
    mock_callback_default_2.assert_called_once_with(event)
    mock_callback_int.assert_called_once_with(event)


@pytest.mark.asyncio
async def test_emitter_unsubscribe_default_callback(event_emitter):
    mock_callback_1 = AsyncMock()
    mock_callback_2 = AsyncMock()

    # Subscribe both callbacks
    event_emitter.subscribe(callback=mock_callback_1)
    event_emitter.subscribe(callback=mock_callback_2)

    # Unsubscribe one callback
    event_emitter.unsubscribe(mock_callback_1)

    event = IntEvent(data=42)

    await event_emitter.emit(event)

    # Check that only the second callback was called
    mock_callback_1.assert_not_called()
    mock_callback_2.assert_called_once_with(event)


@pytest.mark.asyncio
async def test_emitter_unsubscribe_specific_callback(event_emitter):
    mock_callback_int = AsyncMock()

    # Subscribe to IntEvent
    event_emitter.subscribe(IntEvent, callback=mock_callback_int)

    # Unsubscribe
    event_emitter.unsubscribe(mock_callback_int)

    event = IntEvent(data=42)

    await event_emitter.emit(event)

    # Check that the callback was not called
    mock_callback_int.assert_not_called()


@pytest.mark.asyncio
async def test_emitter_unsubscribe_nonexistent_callback(event_emitter):
    mock_callback = AsyncMock()

    # Attempt to unsubscribe a callback that was never subscribed
    # Should not raise an exception
    event_emitter.unsubscribe(mock_callback)

    event = IntEvent(data=42)

    await event_emitter.emit(event)

    # Just a sanity check
    mock_callback.assert_not_called()


@pytest.mark.asyncio
async def test_emitter_emit_no_subscribers(event_emitter):
    # No subscribers registered
    event = IntEvent(data=42)

    # Should not raise any exceptions
    await event_emitter.emit(event)
    # Test passes if no exception is raised


@pytest.mark.asyncio
async def test_emitter_subscribe_same_callback_multiple_times(event_emitter):
    mock_callback = AsyncMock()

    # Subscribe the same callback twice to default events
    event_emitter.subscribe(callback=mock_callback)
    event_emitter.subscribe(callback=mock_callback)

    event = IntEvent(data=42)

    await event_emitter.emit(event)

    # Check that the callback was called exactly once
    # This tests that the emitter uses a set for subscribers to avoid duplicates
    mock_callback.assert_called_once_with(event)


@pytest.mark.asyncio
async def test_emitter_subscribe_same_callback_to_different_events(event_emitter):
    mock_callback = AsyncMock()

    # Subscribe the same callback to both IntEvent and StrEvent
    event_emitter.subscribe(IntEvent, callback=mock_callback)
    event_emitter.subscribe(StrEvent, callback=mock_callback)

    int_event = IntEvent(data=42)
    str_event = StrEvent(data="test")

    await event_emitter.emit(int_event)
    await event_emitter.emit(str_event)

    # Check that the callback was called exactly twice
    assert mock_callback.call_count == 2
    mock_callback.assert_any_call(int_event)
    mock_callback.assert_any_call(str_event)


@pytest.mark.asyncio
async def test_emitter_async_execution_without_await(event_emitter):
    mock_callback = AsyncMock()

    async def mock_callback_async(x):
        await asyncio.sleep(0.1)

    # Add a small delay to the mock to simulate work
    mock_callback.side_effect = mock_callback_async

    event_emitter.subscribe(callback=mock_callback)
    event = IntEvent(data=42)

    # Start the emit but don't await it
    event_emitter.emit(event)
    # This should return immediately without waiting for callbacks

    # Initially the callback might not have been called yet
    assert mock_callback.call_count <= 1

    # Wait for the background task to complete
    await asyncio.sleep(0.5)

    # Now the callback should have been called
    mock_callback.assert_called_once_with(event)



from abc import ABC, abstractmethod
from typing import Any, Dict, Generic, Type

from pydantic import BaseModel, PrivateAttr

from livechain.graph.types import TState


class BaseStatePersister(BaseModel, Generic[TState], ABC):
    state_schema: Type[TState]

    _cached_state: TState | None = PrivateAttr(default=None)

    @abstractmethod
    def _get(self) -> TState:
        raise NotImplementedError

    @abstractmethod
    def _set(self, state: TState | Dict[str, Any]) -> TState:
        raise NotImplementedError

    def get(self) -> TState:
        if self._cached_state is None:
            self._cached_state = self._get()
        return self._cached_state

    def set(self, state: TState | Dict[str, Any]) -> TState:
        self._cached_state = self._set(state)
        return self._cached_state



import asyncio
from unittest.mock import AsyncMock, create_autospec

import pytest
from langgraph.checkpoint.base import BaseCheckpointSaver
from langgraph.store.base import BaseStore
from langgraph.types import RetryPolicy
from pydantic import BaseModel, Field

from livechain.graph.cron import interval
from livechain.graph.executor import Workflow
from livechain.graph.func import cron, reactive, root, step, subscribe
from livechain.graph.ops import get_state, mutate_state, trigger_workflow
from livechain.graph.types import EventSignal, TriggerSignal


class MockState(BaseModel):
    count: int = Field(default=0)


class MockConfig(BaseModel):
    name: str


class MockEvent(EventSignal):
    name: str


@pytest.fixture
def mock_checkpointer():
    checkpointer = create_autospec(BaseCheckpointSaver)
    return checkpointer


@pytest.fixture
def mock_store():
    store = create_autospec(BaseStore)
    return store


@pytest.fixture
def simple_workflow():
    @step()
    async def step_1():
        pass

    @root()
    async def entrypoint():
        pass

    return Workflow.from_routines(entrypoint)


def test_executor_init(simple_workflow):
    simple_workflow.compile(state_schema=MockState)


@pytest.mark.asyncio
async def test_executor_start_validates_input(simple_workflow, mock_checkpointer, mock_store):
    def compile():
        return simple_workflow.compile(
            state_schema=MockState,
            checkpointer=mock_checkpointer,
            store=mock_store,
            config_schema=MockConfig,
        )

    with pytest.raises(ValueError, match="Thread ID is required when using a checkpointer or store"):
        compile().start()

    with pytest.raises(ValueError, match="Config is required when using a config schema"):
        compile().start(thread_id="1")

    with pytest.raises(
        ValueError,
        match="validation error for MockConfig",
    ):
        compile().start(thread_id="1", config={"a": "test"})

    # Should not raise an error
    compile().start(thread_id="1", config={"name": "test"})

    # Should not raise an error
    compile().start(thread_id="1", config=MockConfig(name="test"))


@pytest.mark.asyncio
async def test_executor_basic_workflow_invoked():
    called = False

    @root()
    async def entrypoint():
        step_1()

    @step()
    async def step_1():
        nonlocal called
        called = True

    workflow = Workflow.from_routines(entrypoint)
    executor = workflow.compile(state_schema=MockState)

    executor.start(thread_id="1", config=MockConfig(name="test"))

    assert not called, "Step 1 should not have been called"

    await executor.trigger_workflow(TriggerSignal())

    assert called, "Step 1 should have been called"


@pytest.mark.asyncio
async def test_executor_single_event_routine():
    event_callback = AsyncMock()
    workflow_callback = AsyncMock()

    @root()
    async def entrypoint():
        await workflow_callback()

    @subscribe(event_schema=MockEvent)
    async def on_mock_event(event: MockEvent):
        await event_callback(event)

    workflow = Workflow.from_routines(entrypoint, [on_mock_event])
    executor = workflow.compile(state_schema=MockState)
    executor.start()

    await executor.publish_event(MockEvent(name="test"))

    event_callback.assert_called_once_with(MockEvent(name="test"))
    workflow_callback.assert_not_called()


@pytest.mark.asyncio
async def test_executor_single_reactive_routine():
    reactive_callback = AsyncMock()

    @root()
    async def entrypoint():
        state = get_state(MockState)
        await mutate_state(count=min(1, state.count + 1))

    @reactive(state_schema=MockState, cond=lambda state: state.count)
    async def on_count_change(old_state: MockState, new_state: MockState):
        await reactive_callback(old_state, new_state)

    workflow = Workflow.from_routines(entrypoint, [on_count_change])
    executor = workflow.compile(state_schema=MockState)
    executor.start()

    await executor.trigger_workflow(TriggerSignal())
    await asyncio.sleep(0.2)

    reactive_callback.assert_called_once_with(MockState(count=0), MockState(count=1))
    assert executor.get_state().count == 1, "State should have been updated"

    await executor.trigger_workflow(TriggerSignal())
    await asyncio.sleep(0.2)

    assert executor.get_state().count == 1, "State should not have been updated"


@pytest.mark.asyncio
async def test_executor_single_cron_routine():
    cron_callback = AsyncMock()

    @root()
    async def entrypoint():
        await mutate_state(count=1)

    @cron(expr=interval(seconds=0.2))
    async def on_cron():
        await cron_callback()

    workflow = Workflow.from_routines(entrypoint, [on_cron])
    executor = workflow.compile(state_schema=MockState)
    executor.start()

    await asyncio.sleep(0.3)
    cron_callback.assert_called_once()

    await asyncio.sleep(0.2)
    assert cron_callback.call_count == 2, "Cron should have been called twice"


@pytest.mark.asyncio
async def test_concurrent_event_and_cron_handling():
    """Simulates an agent handling async events while processing scheduled tasks"""
    event_counter = 0
    cron_counter = 0

    @root()
    async def entrypoint():
        pass  # Agent doesn't need direct entrypoint logic

    @subscribe(event_schema=MockEvent)
    async def handle_event(event: MockEvent):
        nonlocal event_counter
        event_counter += 1
        await asyncio.sleep(0.01)  # Simulate processing time

    @cron(expr=interval(seconds=0.05))
    async def background_task():
        nonlocal cron_counter
        cron_counter += 1

    workflow = Workflow.from_routines(entrypoint, [handle_event, background_task])
    executor = workflow.compile(state_schema=MockState)
    executor.start()

    # Send 10 concurrent events
    await asyncio.gather(*[executor.publish_event(MockEvent(name=f"event-{i}")) for i in range(10)])

    await asyncio.sleep(0.2)  # Allow time for processing

    assert event_counter == 10, "All events should be processed"
    assert cron_counter >= 3, "Cron should have triggered multiple times"


@pytest.mark.asyncio
async def test_complex_reactive_workflow():
    """Tests a chained reactive workflow simulating decision-making"""
    state_changes = []

    @root()
    async def entrypoint():
        await mutate_state(count=1)

    @reactive(state_schema=MockState, cond=lambda state: state.count >= 1)
    async def handle_low_priority(old_state, new_state):
        state_changes.append("low-priority")
        await mutate_state(priority="low")

    @reactive(state_schema=MockState, cond=lambda state: state.count >= 5)
    async def handle_high_priority(old_state, new_state):
        state_changes.append("high-priority")
        await mutate_state(priority="high")

    workflow = Workflow.from_routines(entrypoint, [handle_low_priority, handle_high_priority])
    executor = workflow.compile(state_schema=MockState)
    executor.start()

    # Initial state change
    await executor.trigger_workflow(TriggerSignal())
    await asyncio.sleep(0.1)
    assert state_changes == ["low-priority"]

    # Chain state changes
    await executor.mutate_state(MockState(count=5))
    await asyncio.sleep(0.1)
    assert state_changes == ["low-priority", "high-priority"]


@pytest.mark.asyncio
async def test_retry_mechanism_for_failed_steps():
    """Tests automatic retries for failed operations"""
    attempts = 0
    retry_policy = RetryPolicy(initial_interval=0.001, retry_on=ValueError)
    callback = AsyncMock()
    unreliable_done = asyncio.Event()
    callback_done = asyncio.Event()

    @root()
    async def entrypoint():
        pass

    @step(retry=retry_policy)
    async def unreliable_step():
        nonlocal attempts
        attempts += 1

        try:
            if attempts < 3:
                raise ValueError("Temporary failure")
        finally:
            if attempts >= 3:
                unreliable_done.set()

    @subscribe(event_schema=MockEvent)
    async def handle_event(event: MockEvent):
        await unreliable_step()
        await callback()
        callback_done.set()

    workflow = Workflow.from_routines(entrypoint, [handle_event])
    executor = workflow.compile(state_schema=MockState)
    executor.start()

    await executor.publish_event(MockEvent(name="retry-test"))
    await unreliable_done.wait()
    await callback_done.wait()

    assert attempts == 3, f"Should retry 3 times before succeeding, but only succeeded {attempts} times"
    callback.assert_called_once()


@pytest.mark.asyncio
async def test_multi_agent_communication():
    """Simulates multiple agents communicating through topics"""
    received_messages = []

    @root()
    async def agent1_entrypoint():
        pass

    @root()
    async def agent2_entrypoint():
        pass

    # Agent 1 sends messages
    @subscribe(event_schema=MockEvent)
    async def agent1_handler(event: MockEvent):
        await agent2_exec.channel_send("agent-comm", f"Processed {event.name}")

    # Agent 2 receives messages
    def agent2_setup(executor):
        @executor.recv("agent-comm")
        async def handle_message(data: str):
            received_messages.append(data)

    # Create two executors
    agent1 = Workflow.from_routines(agent1_entrypoint, [agent1_handler])
    agent1_exec = agent1.compile(state_schema=MockState)

    agent2 = Workflow.from_routines(agent2_entrypoint)
    agent2_exec = agent2.compile(state_schema=MockState)
    agent2_setup(agent2_exec)

    agent1_exec.start()
    agent2_exec.start()

    # Test communication
    await agent1_exec.publish_event(MockEvent(name="test-msg"))
    await asyncio.sleep(0.2)

    assert received_messages == ["Processed test-msg"]


@pytest.mark.asyncio
async def test_high_load_state_mutations():
    """Stress test for rapid state changes"""
    mutation_count = 0

    @root()
    async def entrypoint():
        pass

    @reactive(state_schema=MockState, cond=lambda state: state.count)
    async def count_handler(old_state, new_state):
        nonlocal mutation_count
        mutation_count += 1

    workflow = Workflow.from_routines(entrypoint, [count_handler])
    executor = workflow.compile(state_schema=MockState)
    executor.start()

    # Send 100 rapid mutations
    await asyncio.gather(*[executor.mutate_state(MockState(count=i)) for i in range(1, 101)])

    await asyncio.sleep(1)
    assert mutation_count == 100, "Should handle all state mutations"


@pytest.mark.asyncio
async def test_workflow_interruption_from_event_handler():
    """Test that triggering workflow from event handler interrupts running workflow"""
    execution_states = []

    @step()
    async def step_1():
        nonlocal execution_states
        execution_states.append("started")

        # Simulate long-running workflow
        await asyncio.sleep(0.2)
        execution_states.append("completed")

    @root()
    async def entrypoint():
        await step_1()

    @subscribe(event_schema=MockEvent)
    async def interrupt_handler(event: MockEvent):
        # Trigger new workflow execution
        await trigger_workflow()

    workflow = Workflow.from_routines(entrypoint, [interrupt_handler])
    executor = workflow.compile(state_schema=MockState)
    executor.start()

    # Start initial workflow
    executor.trigger_workflow(TriggerSignal())
    await asyncio.sleep(0.1)
    executor.publish_event(MockEvent(name="test"))
    await asyncio.sleep(0.25)

    # Verify initial workflow was started but not completed
    assert execution_states == ["started", "started", "completed"]



import asyncio
import time
from abc import ABC, abstractmethod
from heapq import heapify, heappop, heappush
from typing import Dict

from pydantic import BaseModel, Field, PrivateAttr

UnixTime = float  # unix time in seconds


def now() -> UnixTime:
    return time.time()


Second = 1
Minute = Second * 60
Hour = Minute * 60
Day = Hour * 24


class CronExpr(BaseModel, ABC):
    @abstractmethod
    def next_tick(self) -> UnixTime:
        pass


class Interval(CronExpr):
    seconds: float = Field(..., gt=0, description="time in seconds")

    def next_tick(self) -> UnixTime:
        return now() + self.seconds


class Linear(CronExpr):
    base_seconds: float = Field(..., ge=0, description="base time in seconds")

    step_seconds: float = Field(..., gt=0, description="step time in seconds")

    max_interval_seconds: float = Field(..., gt=0, description="max interval between ticks in seconds")

    _count: int = PrivateAttr(default=0)

    def next_tick(self) -> UnixTime:
        next_tick = now() + min(
            self.base_seconds + self.step_seconds * self._count,
            self.max_interval_seconds,
        )
        self._count += 1
        return next_tick


class Exp(CronExpr):
    base_seconds: float = Field(..., gt=0, description="base time in seconds")

    exponent: float = Field(..., ge=1, description="exponent")

    max_interval_seconds: float = Field(..., gt=0, description="max interval between ticks in seconds")

    _count: int = PrivateAttr(default=0)

    def next_tick(self) -> UnixTime:
        next_tick = now() + min(
            self.base_seconds * (self.exponent**self._count),
            self.max_interval_seconds,
        )
        self._count += 1
        return next_tick


class CronJobScheduler(BaseModel):
    cron_jobs: Dict[str, CronExpr]

    async def schedule(self):
        if len(self.cron_jobs) == 0:
            return

        cron_job_pq = [(cron_expr.next_tick(), cron_id, cron_expr) for cron_id, cron_expr in self.cron_jobs.items()]
        heapify(cron_job_pq)

        while True:
            time = now()

            if cron_job_pq[0][0] <= time:
                _, cron_id, cron_expr = heappop(cron_job_pq)
                yield cron_id
                heappush(cron_job_pq, (cron_expr.next_tick(), cron_id, cron_expr))
            else:
                await asyncio.sleep(cron_job_pq[0][0] - time)


def interval(seconds: float) -> CronExpr:
    return Interval(seconds=seconds)


def linear(base: float, step: float, max_interval: float) -> CronExpr:
    return Linear(
        base_seconds=base,
        step_seconds=step,
        max_interval_seconds=max_interval,
    )


def exp(base: float, exponent: float, max_interval: float) -> CronExpr:
    return Exp(
        base_seconds=base,
        exponent=exponent,
        max_interval_seconds=max_interval,
    )



from typing import Any, Awaitable, Callable, Dict

from langchain_core.runnables import Runnable, RunnableLambda, RunnableParallel
from langgraph.pregel.call import SyncAsyncFuture

from livechain.graph.func import step
from livechain.graph.types import P, T


def wrap_in_step(func: Callable[P, Awaitable[T]]) -> Callable[P, SyncAsyncFuture[T]]:
    return step()(func)


def async_parallel(
    func_map: Dict[str, Callable[[T], Awaitable[Any]]],
) -> Runnable[T, Dict[str, Any]]:
    f = {k: RunnableLambda(func) for k, func in func_map.items()}
    runnable = RunnableParallel[T](**f)  # type: ignore
    return runnable


def rename_function(new_name: str):
    def decorator(func: Callable[P, T]):
        func.__name__ = new_name
        return func

    return decorator


# def step_gather(
#     *funcs: Callable[P, Awaitable[T]],
# ) -> Callable[P, SyncAsyncFuture[List[T]]]:
#     substeps = [wrap_in_step(func) for func in funcs]

#     @step(name="gather")
#     async def gather_step(*args: P.args, **kwargs: P.kwargs) -> List[Any]:
#         return await asyncio.gather(
#             *[substep(*args, **kwargs) for substep in substeps],
#             return_exceptions=False,
#         )

#     return gather_step



from time import sleep

from pydantic import BaseModel

from livechain.graph import EventSignal, Workflow, cron, reactive, root, step, subscribe
from livechain.graph.cron import interval


class MyEvent(EventSignal):
    data: int


class MyState(BaseModel):
    data: int


@root()
async def root_fn():
    pass


@step()
async def step_fn():
    pass


@subscribe(MyEvent)
async def event_signal_fn(event: MyEvent):
    pass


@reactive(MyState, cond=lambda state: state.data > 10)
async def reactive_fn(old_state: MyState, new_state: MyState):
    pass


@cron(expr=interval(seconds=1))
async def cron_fn():
    pass


wf = Workflow.from_routines(root_fn, [event_signal_fn, reactive_fn, cron_fn])
executor = wf.compile(MyState)

executor.start()
sleep(1)
executor.stop()



import asyncio
from typing import Any, Awaitable, Callable, Dict, Optional

import pytest
from pydantic import BaseModel

from livechain.graph.cron import interval
from livechain.graph.func import cron, reactive, step, subscribe
from livechain.graph.func.routine import (
    CronSignalRoutine,
    EventSignalRoutine,
    Mode,
    ReactiveSignalRoutine,
    SignalRoutineType,
    SignalStrategy,
)
from livechain.graph.types import CronSignal, EventSignal, ReactiveSignal
from livechain.graph.utils import run_in_context


# Sample models for testing
class MockEvent(EventSignal):
    name: str
    value: int


class MockState(BaseModel):
    count: int
    message: str
    data: Optional[Dict[str, Any]] = None


def create_cron_routine(mode: SignalStrategy, func: Callable[[], Awaitable[Any]]):
    @cron(interval(seconds=60), strategy=mode)
    async def test_cron():
        return await func()

    return test_cron


def create_reactive_routine(mode: SignalStrategy, func: Callable[[MockState, MockState], Awaitable[Any]]):
    @reactive(MockState, lambda state: state.count, strategy=mode)
    async def test_reactive(old_state: MockState, new_state: MockState):
        return await func(old_state, new_state)

    return test_reactive


def create_event_routine(mode: SignalStrategy, func: Callable[[MockEvent], Awaitable[Any]]):
    @subscribe(MockEvent, strategy=mode)
    async def test_subscriber(event: MockEvent):
        return await func(event)

    return test_subscriber


def test_subscribe_decorator():
    """Test the subscribe decorator creates an EventSignalRoutine."""

    @subscribe(MockEvent, strategy=Mode.Parallel())
    async def test_subscriber(event: MockEvent):
        pass

    # Verify the result is an EventSignalRoutine with correct properties
    assert isinstance(test_subscriber, EventSignalRoutine)
    assert test_subscriber.routine_type == SignalRoutineType.SUBSCRIBE
    assert test_subscriber.schema == MockEvent
    assert test_subscriber.name == "test_subscriber"
    assert isinstance(test_subscriber.mode, Mode.Parallel)


def test_subscribe_decorator_with_name():
    """Test the subscribe decorator creates an EventSignalRoutine with a name."""

    @subscribe(MockEvent, strategy=Mode.Parallel(), name="override_name")
    async def test_subscriber(event: MockEvent):
        pass

    assert test_subscriber.name == "override_name"


@pytest.mark.asyncio
async def test_subscribe_runnable_execution():
    """Test the subscribe decorator function's execution flow."""
    # Define the subscriber function
    subscriber_calls = []

    @subscribe(MockEvent, strategy=Mode.Queue(), name="test_subscriber")
    async def test_subscriber(event: MockEvent):
        subscriber_calls.append(event)

    # Create a runnable and invoke it
    runnable = test_subscriber.create_routine_runnable()
    test_event = MockEvent(name="test", value=42)
    await runnable.ainvoke(test_event)

    # Verify the event was passed to the subscriber function
    assert len(subscriber_calls) == 1
    assert subscriber_calls[0] == test_event


def test_reactive_decorator():
    """Test the reactive decorator creates a ReactiveSignalRoutine."""

    # Define a condition to watch for changes
    def watched_value(state):
        return state.count

    @reactive(MockState, watched_value, strategy=Mode.Interrupt())
    async def test_reactive(old_state: MockState, new_state: MockState):
        pass

    # Verify the result is a ReactiveSignalRoutine with correct properties
    assert isinstance(test_reactive, ReactiveSignalRoutine)
    assert test_reactive.routine_type == SignalRoutineType.REACTIVE
    assert test_reactive.schema == ReactiveSignal[MockState]
    assert test_reactive.state_schema == MockState
    assert test_reactive.name == "test_reactive"
    assert test_reactive.cond == watched_value
    assert isinstance(test_reactive.mode, Mode.Interrupt)


@pytest.mark.asyncio
async def test_reactive_execution():
    """Test the reactive decorator function's execution flow."""

    # Define a condition to watch for changes
    def watched_value(state):
        return state.count

    # Define the reactive effect
    effect_calls = []

    @reactive(MockState, watched_value, strategy=Mode.Parallel(), name="test_reactive")
    async def test_reactive(old_state: MockState, new_state: MockState):
        effect_calls.append((old_state, new_state))

    # Create a runnable and invoke it
    runnable = test_reactive.create_routine_runnable()

    old_state = MockState(count=1, message="old message")
    new_state = MockState(count=2, message="new message")

    signal = ReactiveSignal(old_state=old_state, new_state=new_state)
    await runnable.ainvoke(signal)

    # Verify the states were passed to the effect function
    assert len(effect_calls) == 1
    assert effect_calls[0][0] == old_state
    assert effect_calls[0][1] == new_state


def test_cron_decorator():
    """Test the cron decorator creates a CronSignalRoutine."""
    # Define a cron expression
    cron_expr = interval(seconds=60)

    @cron(cron_expr, strategy=Mode.Parallel())
    async def test_cron():
        pass

    # Verify the result is a CronSignalRoutine with correct properties
    assert isinstance(test_cron, CronSignalRoutine)
    assert test_cron.routine_type == SignalRoutineType.CRON
    assert test_cron.schema == CronSignal
    assert test_cron.name == "test_cron"
    assert test_cron.cron_expr == cron_expr
    assert isinstance(test_cron.mode, Mode.Parallel)


@pytest.mark.asyncio
async def test_cron_execution():
    """Test the cron decorator function's execution flow."""
    # Define a cron expression
    cron_expr = interval(seconds=30)

    # Define the cron effect
    effect_calls = []

    @cron(cron_expr, strategy=Mode.Queue(), name="test_cron")
    async def test_cron():
        effect_calls.append("cron_triggered")
        return "cron_result"

    # Create a runnable and invoke it
    runnable = test_cron.create_routine_runnable()

    signal = CronSignal(cron_id="test_cron_job")
    await runnable.ainvoke(signal)

    # Verify the effect was called
    assert len(effect_calls) == 1
    assert effect_calls[0] == "cron_triggered"


@pytest.mark.asyncio
async def test_subscribe_multiple_modes():
    """Test the subscribe decorator with multiple modes."""

    async def test_subscriber(event: MockEvent):
        pass

    sub_parallel = create_event_routine(Mode.Parallel(), test_subscriber)
    sub_queue = create_event_routine(Mode.Queue(), test_subscriber)
    create_event_routine(Mode.Interrupt(), test_subscriber)

    sub_parallel.create_routine_runnable()
    sub_queue.create_routine_runnable()


@pytest.mark.asyncio
async def test_runner_behavior_interrupt_mode():
    """Test that interrupt mode cancels previous tasks."""
    # Track execution
    exec_history = []
    done = asyncio.Event()

    # Create a mock function with controlled execution time
    async def mock_handler(event: MockEvent):
        exec_history.append(f"start-{event.name}")
        if event.name == "long":
            # Long-running task
            await asyncio.sleep(0.2)
        else:
            # Short task
            await asyncio.sleep(0.05)
        exec_history.append(f"end-{event.name}")

        if event.name == "last":
            done.set()

    # Create routine with interrupt mode
    routine = create_event_routine(Mode.Interrupt(), mock_handler)

    # Create the runner
    runner = routine.create_runner()

    # Start the runner in the background
    runner_task = asyncio.create_task(runner.start())

    # Send a long-running task followed quickly by a short task
    await runner(MockEvent(name="long", value=1))
    await asyncio.sleep(0.05)  # Wait a bit but not enough for long task to complete
    await runner(MockEvent(name="short", value=2))
    await asyncio.sleep(0.1)  # Wait for short task to complete
    await runner(MockEvent(name="last", value=3))

    # Wait for all to complete
    await done.wait()

    # Stop the runner
    runner.stop()
    await runner_task

    # Order should be: start-long, start-short, end-short, start-last, end-last
    assert exec_history == [
        "start-long",
        "start-short",
        "end-short",
        "start-last",
        "end-last",
    ]


@pytest.mark.asyncio
async def test_runner_behavior_parallel_mode():
    """Test that parallel mode runs tasks concurrently."""
    # Track execution
    exec_history = []
    done = asyncio.Event()

    # Create a mock function with controlled execution time
    async def mock_handler(event: MockEvent):
        exec_history.append(f"start-{event.name}")
        # Simulate work
        await asyncio.sleep(0.1)
        exec_history.append(f"end-{event.name}")

        if event.name == "third":
            done.set()

    # Create routine with parallel mode
    routine = create_event_routine(Mode.Parallel(), mock_handler)

    # Create the runner
    runner = routine.create_runner()

    # Start the runner in the background
    runner_task = asyncio.create_task(runner.start())

    # Send multiple events in quick succession
    await runner(MockEvent(name="first", value=1))
    await asyncio.sleep(0.01)
    await runner(MockEvent(name="second", value=2))
    await asyncio.sleep(0.01)
    await runner(MockEvent(name="third", value=3))

    # Wait for all to complete
    await done.wait()

    # Stop the runner
    runner.stop()
    await runner_task

    # Verify behavior: all tasks should start before any completes
    starts = [i for i in exec_history if i.startswith("start-")]
    ends = [i for i in exec_history if i.startswith("end-")]

    # All tasks should be started and completed
    assert len(starts) == 3
    assert len(ends) == 3

    # All starts should happen before any ends
    # (this is true because our sleep time is the same for all tasks)
    assert all(exec_history.index(start) < exec_history.index(f"end-{start[6:]}") for start in starts)

    # The starts should be in order (first, second, third)
    # because we send them sequentially
    assert starts == ["start-first", "start-second", "start-third"]


@pytest.mark.asyncio
async def test_runner_behavior_queue_mode():
    """Test that queue mode processes tasks in order."""
    # Track execution
    exec_history = []
    done = asyncio.Event()

    # Create a mock function with controlled execution time
    async def mock_handler(event: MockEvent):
        exec_history.append(f"start-{event.name}")
        # Simulate work with different durations
        await asyncio.sleep(0.1 if event.name == "second" else 0.05)
        exec_history.append(f"end-{event.name}")

        if event.name == "third":
            done.set()

    # Create routine with queue mode
    routine = create_event_routine(Mode.Queue(), mock_handler)

    # Create the runner
    runner = routine.create_runner()

    # Start the runner in the background
    runner_task = asyncio.create_task(runner.start())

    # Send multiple events in quick succession
    await runner(MockEvent(name="first", value=1))
    await asyncio.sleep(0.01)
    await runner(MockEvent(name="second", value=2))
    await asyncio.sleep(0.01)
    await runner(MockEvent(name="third", value=3))

    # Wait for all to complete
    await done.wait()

    # Stop the runner
    runner.stop()
    await runner_task

    # Verify behavior: tasks should be processed in order (FIFO)
    # Order should be: start-first, end-first, start-second, end-second, start-third, end-third
    expected_order = [
        "start-first",
        "end-first",
        "start-second",
        "end-second",
        "start-third",
        "end-third",
    ]

    assert exec_history == expected_order


@pytest.mark.asyncio
async def test_runner_behavior_debounce_mode():
    """Test that debounce mode only processes the last task after delay."""
    # Track execution
    exec_history = []
    done = asyncio.Event()

    # Create a mock function
    async def mock_handler(event: MockEvent):
        exec_history.append(f"executed-{event.name}")

        if event.name == "fourth":
            done.set()

    # Create debounce mode with 0.1s delay
    debounce_mode = Mode.Debounce(delay=0.1)

    # Create routine with debounce mode
    routine = create_event_routine(debounce_mode, mock_handler)

    # Create the runner
    runner = routine.create_runner()

    # Start the runner in the background
    runner_task = asyncio.create_task(runner.start())

    # Send multiple events within the debounce window
    await runner(MockEvent(name="first", value=1))
    await asyncio.sleep(0.05)  # Less than debounce delay
    await runner(MockEvent(name="second", value=2))
    await asyncio.sleep(0.05)  # Still less than full delay from second event
    await runner(MockEvent(name="third", value=3))
    await asyncio.sleep(0.2)  # Wait for debounce delay to pass
    await runner(MockEvent(name="fourth", value=4))

    # Wait for debounce delay to pass
    await done.wait()

    # Stop the runner
    runner.stop()
    await runner_task

    # Verify behavior: only the last two events in each debounce window should be executed
    assert exec_history == ["executed-third", "executed-fourth"]


@pytest.mark.asyncio
async def test_step_preserves_function_name():
    """Test that step decorator preserves the original function name."""

    @step()
    async def my_test_function(x):
        return x + 1

    # Verify the function name is preserved
    assert my_test_function.__name__ == "my_test_function"

    @run_in_context
    async def test_step_with_context():
        assert await my_test_function(5) == 6

    await test_step_with_context()


def test_step_requires_async_function():
    """Test that step decorator raises an error when used with non-async function."""

    # This should raise a TypeError or ValueError
    with pytest.raises((TypeError, ValueError)):

        @step()  # type: ignore
        def non_async_function(data):
            return f"processed_{data}"


if __name__ == "__main__":
    pytest.main()



from inspect import iscoroutinefunction
from typing import Any, Awaitable, Callable, Dict, Optional, overload

from langchain_core.callbacks.base import Callbacks
from langchain_core.runnables import RunnableConfig
from langgraph.func import entrypoint
from pydantic import BaseModel

from livechain.graph.constants import CONF, CONFIG_KEY_CONTEXT
from livechain.graph.context import Context


def make_config(configurable: Dict[str, Any], callbacks: Callbacks = None) -> RunnableConfig:
    return {CONF: configurable, "callbacks": callbacks}


def make_config_from_context(
    context: Context,
    thread_id: Optional[str] = None,
    config: Optional[Dict[str, Any] | BaseModel] = None,
    callbacks: Callbacks = None,
) -> RunnableConfig:
    if isinstance(config, BaseModel):
        config = config.model_dump()

    configurable: Dict[str, Any] = {CONFIG_KEY_CONTEXT: context}

    if thread_id is not None:
        configurable["thread_id"] = thread_id

    if config is not None:
        configurable.update(config)

    return make_config(configurable, callbacks)


@overload
def run_in_context(func: Callable[..., None]) -> Callable[..., None]: ...


@overload
def run_in_context(
    func: Callable[..., Awaitable[None]],
) -> Callable[..., Awaitable[None]]: ...


def run_in_context(
    func: Callable[..., Any] | Callable[..., Awaitable[Any]],
) -> Callable[..., Any] | Callable[..., Awaitable[Any]]:
    @entrypoint()
    async def run_async_in_context_wrapper(input: Any):
        return await func()

    @entrypoint()
    def run_sync_in_context_wrapper(input: Any):
        return func()

    if iscoroutinefunction(func):
        pregel = run_async_in_context_wrapper
    else:
        pregel = run_sync_in_context_wrapper

    async def arun_entrypoint():
        return await pregel.ainvoke(1)

    def run_entrypoint():
        return pregel.invoke(1)

    if iscoroutinefunction(func):
        return arun_entrypoint
    else:
        return run_entrypoint



from unittest.mock import patch

import pytest

from livechain.graph.cron import CronJobScheduler, Exp, Interval, Linear, exp, interval, linear


class TestCronExpressions:
    """Test the different CronExpr implementations."""

    def test_interval(self):
        """Test that Interval correctly calculates the next tick."""
        with patch("livechain.graph.cron.now", return_value=100.0):
            interval_expr = Interval(seconds=30)
            assert interval_expr.next_tick() == 130.0

    def test_linear(self):
        """Test that Linear correctly calculates next ticks with increasing intervals."""
        with patch("livechain.graph.cron.now", return_value=100.0):
            linear_expr = Linear(base_seconds=10, step_seconds=5, max_interval_seconds=30)

            # First call: base + (step * 0) = 10 + 0 = 10
            assert linear_expr.next_tick() == 110.0

            # Second call: base + (step * 1) = 10 + 5 = 15
            assert linear_expr.next_tick() == 115.0

            # Third call: base + (step * 2) = 10 + 10 = 20
            assert linear_expr.next_tick() == 120.0

            # Let's skip to where we hit the max
            linear_expr._count = 5  # base + (step * 5) = 10 + 25 = 35 > max(30)
            assert linear_expr.next_tick() == 130.0  # Should use max_interval

    def test_exp(self):
        """Test that Exp correctly calculates next ticks with exponentially increasing intervals."""
        with patch("livechain.graph.cron.now", return_value=100.0):
            exp_expr = Exp(base_seconds=2, exponent=2, max_interval_seconds=20)

            # First call: base * (exponent ** 0) = 2 * 1 = 2
            assert exp_expr.next_tick() == 102.0

            # Second call: base * (exponent ** 1) = 2 * 2 = 4
            assert exp_expr.next_tick() == 104.0

            # Third call: base * (exponent ** 2) = 2 * 4 = 8
            assert exp_expr.next_tick() == 108.0

            # Fourth call: base * (exponent ** 3) = 2 * 8 = 16
            assert exp_expr.next_tick() == 116.0

            # Fifth call: base * (exponent ** 4) = 2 * 16 = 32 > max(20)
            assert exp_expr.next_tick() == 120.0  # Should use max_interval

    def test_helper_functions(self):
        """Test the helper functions for creating cron expressions."""
        assert isinstance(interval(30), Interval)
        assert isinstance(linear(10, 5, 30), Linear)
        assert isinstance(exp(2, 2, 20), Exp)


class TestCronJobScheduler:
    """Test the CronJobScheduler class."""

    @pytest.mark.asyncio
    async def test_scheduler_no_jobs(self):
        """Test that scheduler with no jobs does nothing."""
        scheduler = CronJobScheduler(cron_jobs={})
        async for _job in scheduler.schedule():
            raise AssertionError()

    @pytest.mark.asyncio
    async def test_scheduler_basic_intervals(self):
        """Test that scheduler with mock sleep that advances time."""
        # Control time with a mutable variable
        current_time = 100.0

        def mock_time():
            nonlocal current_time
            return current_time

        async def mock_sleep(seconds):
            nonlocal current_time
            # Advance time when sleep is called
            current_time += seconds

        with (
            patch("livechain.graph.cron.now", side_effect=mock_time),
            patch("asyncio.sleep", side_effect=mock_sleep),
        ):
            # Create scheduler with two jobs with different intervals
            scheduler = CronJobScheduler(
                cron_jobs={
                    "job1": Interval(seconds=15),  # Next tick at 115
                    "job2": Interval(seconds=25),  # Next tick at 125
                }
            )

            # Run the scheduler
            results = []

            i = 0
            async for job in scheduler.schedule():
                results.append((job, current_time))
                i += 1
                if i >= 5:
                    break

            # Verify jobs were executed at expected times
            assert results == [
                ("job1", 115.0),  # job1 first tick
                ("job2", 125.0),  # job2 first tick
                ("job1", 130.0),  # job1 second tick
                ("job1", 145.0),  # job1 third tick
                ("job2", 150.0),  # job2 fourth tick
            ]

    @pytest.mark.asyncio
    async def test_scheduler_with_same_intervals(self):
        """Test scheduler with multiple jobs that have the same interval."""
        current_time = 100.0

        def mock_time():
            nonlocal current_time
            return current_time

        async def mock_sleep(seconds):
            nonlocal current_time
            current_time += seconds

        with (
            patch("livechain.graph.cron.now", side_effect=mock_time),
            patch("asyncio.sleep", side_effect=mock_sleep),
        ):
            # Create scheduler with jobs having identical intervals
            scheduler = CronJobScheduler(
                cron_jobs={
                    "job1": Interval(seconds=20),  # Next tick at 120
                    "job2": Interval(seconds=20),  # Next tick at 120
                }
            )

            # Run the scheduler
            results = []
            i = 0
            async for job in scheduler.schedule():
                results.append((job, current_time))
                i += 1
                if i >= 4:
                    break

            # Jobs with same interval should alternate in their execution
            # But the initial order depends on heap insertion order
            assert len(results) == 4

            # Both jobs should run at the same timestamps
            assert results[0][1] == 120.0
            assert results[1][1] == 120.0
            assert results[2][1] == 140.0
            assert results[3][1] == 140.0

            # Check that both jobs executed twice
            assert results[0][0] != results[1][0]
            assert results[2][0] != results[3][0]
            assert {results[0][0], results[1][0]} == {"job1", "job2"}
            assert {results[2][0], results[3][0]} == {"job1", "job2"}

    @pytest.mark.asyncio
    async def test_scheduler_with_linear_cron(self):
        """Test scheduler with Linear cron expressions."""
        current_time = 100.0

        def mock_time():
            nonlocal current_time
            return current_time

        async def mock_sleep(seconds):
            nonlocal current_time
            current_time += seconds

        with (
            patch("livechain.graph.cron.now", side_effect=mock_time),
            patch("asyncio.sleep", side_effect=mock_sleep),
        ):
            # Create scheduler with a Linear cron expression
            # base=5, step=5, max=20
            # Intervals: 5, 10, 15, 20, 20, ...
            scheduler = CronJobScheduler(
                cron_jobs={
                    "linear_job": Linear(base_seconds=5, step_seconds=5, max_interval_seconds=20),
                }
            )

            # Run the scheduler
            results = []
            i = 0
            async for job in scheduler.schedule():
                results.append((job, current_time))
                i += 1
                if i >= 5:
                    break

            # Verify jobs were executed at the expected times with increasing intervals
            assert results == [
                ("linear_job", 105.0),  # First tick: base=5
                ("linear_job", 115.0),  # Second tick: base + step = 10
                ("linear_job", 130.0),  # Third tick: base + 2*step = 15
                ("linear_job", 150.0),  # Fourth tick: base + 3*step = 20
                ("linear_job", 170.0),  # Fifth tick: max = 20
            ]

    @pytest.mark.asyncio
    async def test_scheduler_with_exp_cron(self):
        """Test scheduler with Exponential cron expressions."""
        current_time = 100.0

        def mock_time():
            nonlocal current_time
            return current_time

        async def mock_sleep(seconds):
            nonlocal current_time
            current_time += seconds

        with (
            patch("livechain.graph.cron.now", side_effect=mock_time),
            patch("asyncio.sleep", side_effect=mock_sleep),
        ):
            # Create scheduler with an Exp cron expression
            # base=2, exponent=2, max=20
            # Intervals: 2, 4, 8, 16, 20, 20, ...
            scheduler = CronJobScheduler(
                cron_jobs={
                    "exp_job": Exp(base_seconds=2, exponent=2, max_interval_seconds=20),
                }
            )

            # Run the scheduler
            results = []
            i = 0
            async for job in scheduler.schedule():
                results.append((job, current_time))
                i += 1
                if i >= 5:
                    break

            # Verify jobs were executed at the expected times with exponentially increasing intervals
            assert results == [
                ("exp_job", 102.0),  # First tick: base=2
                ("exp_job", 106.0),  # Second tick: base * exponent^1 = 4
                ("exp_job", 114.0),  # Third tick: base * exponent^2 = 8
                ("exp_job", 130.0),  # Fourth tick: base * exponent^3 = 16
                ("exp_job", 150.0),  # Fifth tick: max = 20
            ]

    @pytest.mark.asyncio
    async def test_scheduler_mixed_cron_types(self):
        """Test scheduler with a mix of different cron expression types."""
        current_time = 100.0

        def mock_time():
            nonlocal current_time
            return current_time

        async def mock_sleep(seconds):
            nonlocal current_time
            current_time += seconds

        with (
            patch("livechain.graph.cron.now", side_effect=mock_time),
            patch("asyncio.sleep", side_effect=mock_sleep),
        ):
            # Create scheduler with different types of cron expressions
            scheduler = CronJobScheduler(
                cron_jobs={
                    "interval_job": Interval(seconds=8),
                    "linear_job": Linear(base_seconds=5, step_seconds=5, max_interval_seconds=30),
                    "exp_job": Exp(base_seconds=3, exponent=2, max_interval_seconds=25),
                }
            )

            # Run the scheduler
            results = []
            i = 0
            async for job in scheduler.schedule():
                results.append((job, current_time))
                i += 1
                if i >= 8:
                    break

            # First 6 jobs should be in this order and at these times:
            # interval_job at 108, 116, 124, 132, 140, 148
            # linear_job at 105, 115, 130
            # exp_job at 103, 109, 121, 125, 125
            assert results == [
                ("exp_job", 103.0),
                ("linear_job", 105.0),
                ("interval_job", 108.0),
                ("exp_job", 109.0),
                ("linear_job", 115.0),
                ("interval_job", 116.0),
                ("exp_job", 121.0),
                ("interval_job", 124.0),
            ]


if __name__ == "__main__":
    pytest.main()



from __future__ import annotations

from typing import Any, Awaitable, Callable, Generic, Hashable, Optional, Protocol, Type, TypeVar

from langgraph.checkpoint.base import BaseCheckpointSaver
from langgraph.store.base import BaseStore
from pydantic import BaseModel, ConfigDict
from typing_extensions import ParamSpec

TState = TypeVar("TState", bound=BaseModel)
TConfig = TypeVar("TConfig", bound=BaseModel)
TTopic = TypeVar("TTopic", bound=str)
TModel = TypeVar("TModel", bound=BaseModel)
THashable = TypeVar("THashable", bound=Hashable)

P = ParamSpec("P")
T = TypeVar("T")


TState_contra = TypeVar("TState_contra", bound=BaseModel, contravariant=True)
TModel_contra = TypeVar("TModel_contra", bound=BaseModel, contravariant=True)
T_cov = TypeVar("T_cov", covariant=True)


EntrypointFunc = Callable[[], Awaitable[None]]


class EventSignal(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)


TEvent = TypeVar("TEvent", bound=EventSignal)


class ReactiveSignal(BaseModel, Generic[TState]):
    old_state: TState
    new_state: TState


class CronSignal(BaseModel):
    cron_id: str


class TopicSignal(BaseModel):
    topic: str
    data: Any


class TriggerSignal(BaseModel): ...


class WatchedValue(Protocol, Generic[TState_contra, T_cov]):
    def __call__(self, state: TState_contra) -> T_cov: ...


class Subscriber(Protocol, Generic[TModel_contra]):
    def __call__(self, event: TModel_contra) -> Awaitable[Any]: ...


class ReactiveEffect(Protocol, Generic[TState_contra]):
    def __call__(self, old_state: TState_contra, new_state: TState_contra) -> Awaitable[Any]: ...


class CronEffect(Protocol):
    def __call__(self) -> Awaitable[Any]: ...


class LangGraphInjectable(BaseModel):
    """Injectable dependencies for LangGraph."""

    checkpointer: Optional[BaseCheckpointSaver] = None
    store: Optional[BaseStore] = None
    config_schema: Optional[Type[Any]] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    def from_empty(cls) -> LangGraphInjectable:
        return cls()

    @classmethod
    def from_values(
        cls,
        checkpointer: Optional[BaseCheckpointSaver] = None,
        store: Optional[BaseStore] = None,
        config_schema: Optional[Type[Any]] = None,
    ) -> LangGraphInjectable:
        return cls(checkpointer=checkpointer, store=store, config_schema=config_schema)

    @property
    def require_thread_id(self) -> bool:
        return self.checkpointer is not None or self.store is not None

    @property
    def require_config(self) -> bool:
        return self.config_schema is not None



# Import main modules
# Import version
from livechain.__about__ import __version__  # noqa: F401
from livechain.graph import (  # noqa: F401
    constants,
    context,
    cron,
    emitter,
    executor,
    ops,
    reactive,
    root,
    step,
    subscribe,
    types,
    utils,
)

# Export key functionality
__all__ = [
    "context",
    "constants",
    "cron",
    "emitter",
    "executor",
    "ops",
    "types",
    "utils",
    "root",
    "step",
    "subscribe",
    "reactive",
    "__version__",
]



from __future__ import annotations

import asyncio
import logging
from typing import Any, Awaitable, Callable, Dict, Generic, List, Optional, Type, Union, cast

from langchain_core.callbacks.base import Callbacks
from langgraph.checkpoint.base import BaseCheckpointSaver
from langgraph.store.base import BaseStore
from pydantic import BaseModel, ConfigDict, PrivateAttr

from livechain.graph.context import Context
from livechain.graph.cron import CronExpr, CronJobScheduler
from livechain.graph.func.routine import (
    CronSignalRoutine,
    EventSignalRoutine,
    ReactiveSignalRoutine,
    SignalRoutineRunner,
    SignalRoutineType,
    WorkflowSignalRoutine,
)
from livechain.graph.func.utils import rename_function
from livechain.graph.persist.base import BaseStatePersister
from livechain.graph.types import (
    EventSignal,
    LangGraphInjectable,
    ReactiveSignal,
    TConfig,
    TopicSignal,
    TriggerSignal,
    TState,
    TTopic,
    WatchedValue,
)
from livechain.graph.utils import make_config_from_context

logger = logging.getLogger(__name__)

BackgroundRoutine = Union[EventSignalRoutine, CronSignalRoutine, ReactiveSignalRoutine]


class Workflow(BaseModel, Generic[TState, TConfig, TTopic]):
    root: WorkflowSignalRoutine

    routines: List[BackgroundRoutine]

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    def from_routines(
        cls,
        root: WorkflowSignalRoutine,
        routines: List[BackgroundRoutine] | None = None,
    ) -> Workflow:
        if routines is None:
            routines = []

        return cls(root=root, routines=routines)

    def compile(
        self,
        state_schema: Type[TState],
        persister: Optional[BaseStatePersister[TState]] = None,
        *,
        checkpointer: Optional[BaseCheckpointSaver] = None,
        store: Optional[BaseStore] = None,
        config_schema: Optional[Type[TConfig]] = None,
    ) -> WorkflowExecutor:
        context = Context(state_schema=state_schema, persister=persister)
        injectable = LangGraphInjectable.from_values(
            checkpointer=checkpointer,
            store=store,
            config_schema=config_schema,
        )

        event_routines: List[EventSignalRoutine[EventSignal]] = []
        cron_routines: List[CronSignalRoutine] = []
        reactive_routines: List[ReactiveSignalRoutine[TState, Any]] = []

        for routine in self.routines:
            if routine.routine_type == SignalRoutineType.SUBSCRIBE:
                event_routines.append(cast(EventSignalRoutine, routine))
            elif routine.routine_type == SignalRoutineType.CRON:
                cron_routines.append(cast(CronSignalRoutine, routine))
            elif routine.routine_type == SignalRoutineType.REACTIVE:
                reactive_routines.append(cast(ReactiveSignalRoutine, routine))

        for reactive_routine in reactive_routines:
            if reactive_routine.state_schema != state_schema:
                raise ValueError(
                    f"Reactive routine {reactive_routine.name} has state schema {reactive_routine.state_schema}, "
                    f"which does not match the workflow state schema {state_schema}"
                )

        return WorkflowExecutor(
            injectable=injectable,
            context=context,
            workflow_routine=self.root,
            event_routines=event_routines,
            cron_routines=cron_routines,
            reactive_routines=reactive_routines,
        )


class WorkflowExecutor(BaseModel, Generic[TState, TConfig, TTopic]):
    _injectable: LangGraphInjectable = PrivateAttr()

    _context: Context[TState] = PrivateAttr()

    _workflow_routine: WorkflowSignalRoutine = PrivateAttr()

    _event_routines: List[EventSignalRoutine[EventSignal]] = PrivateAttr()

    _cron_routines: List[CronSignalRoutine] = PrivateAttr()

    _reactive_routines: List[ReactiveSignalRoutine[TState, Any]] = PrivateAttr()

    _workflow_task: Optional[asyncio.Task] = PrivateAttr(default=None)

    _executor_tasks: List[asyncio.Task[None]] = PrivateAttr(default_factory=list)

    _runners: List[SignalRoutineRunner] = PrivateAttr(default_factory=list)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __init__(
        self,
        injectable: LangGraphInjectable,
        context: Context,
        workflow_routine: WorkflowSignalRoutine,
        event_routines: List[EventSignalRoutine[EventSignal]],
        cron_routines: List[CronSignalRoutine],
        reactive_routines: List[ReactiveSignalRoutine[TState, Any]],
    ):
        super().__init__()
        self._injectable = injectable
        self._context = context
        self._workflow_routine = workflow_routine
        self._event_routines = event_routines
        self._cron_routines = cron_routines
        self._reactive_routines = reactive_routines

    def start(
        self,
        thread_id: Optional[str] = None,
        config: Optional[TConfig | Dict[str, Any]] = None,
        callbacks: Callbacks = None,
    ):
        if self._injectable.require_thread_id and thread_id is None:
            raise ValueError("Thread ID is required when using a checkpointer or store")

        if self._injectable.require_config and config is None:
            raise ValueError("Config is required when using a config schema")

        if self._injectable.config_schema is not None:
            validated_config = self._injectable.config_schema.model_validate(config)
        else:
            validated_config = config

        cron_jobs: Dict[str, CronExpr] = {}
        runnable_config = make_config_from_context(self._context, thread_id, validated_config, callbacks)

        for event_routine in self._event_routines:
            routine_runner = event_routine.create_runner(config=runnable_config, injectable=self._injectable)
            self._runners.append(routine_runner)
            self._context.events.subscribe(event_routine.schema, callback=routine_runner)

        for cron_routine in self._cron_routines:
            routine_runner = cron_routine.create_runner(config=runnable_config, injectable=self._injectable)
            cron_jobs[routine_runner.routine_id] = cron_routine.cron_expr
            self._runners.append(routine_runner)
            self._context.cron_jobs.subscribe(routine_runner.routine_id, callback=routine_runner)

        for reactive_routine in self._reactive_routines:
            routine_runner = reactive_routine.create_runner(config=runnable_config, injectable=self._injectable)
            conditional_callback = _with_cond(reactive_routine.cond, routine_runner)
            self._runners.append(routine_runner)
            self._context.effects.subscribe(callback=conditional_callback)

        workflow_runner = self._workflow_routine.create_runner(config=runnable_config, injectable=self._injectable)
        self._runners.append(workflow_runner)

        # register a callback to trigger the main workflow
        self._context.trigger.subscribe(callback=workflow_runner)

        self._executor_tasks = [
            asyncio.create_task(self._schedule_cron_jobs(cron_jobs)),
            *[asyncio.create_task(runner.start()) for runner in self._runners],
        ]

        asyncio.gather(*self._executor_tasks, return_exceptions=False)

    def stop(self):
        logger.info("Stopping workflow")
        for runner in self._runners:
            runner.stop()

        logger.info("Waiting for runners to stop")
        for task in self._executor_tasks:
            task.cancel()

        logger.info("Cancelling workflow task")
        if self._workflow_task is not None:
            self._workflow_task.cancel()

        logger.info("Unsubscribing from events, effects, cron jobs, and trigger")
        self._context.events.unsubscribe_all()
        self._context.effects.unsubscribe_all()
        self._context.cron_jobs.unsubscribe_all()
        self._context.trigger.unsubscribe_all()

        logger.info("Resetting workflow executor inner state")
        self._workflow_task = None
        self._executor_tasks = []
        self._runners = []

    def recv(self, topic: TTopic):
        def recv_decorator(func: Callable[[Any], Awaitable[Any]]):
            async def func_wrapper(signal: TopicSignal):
                return await func(signal.data)

            self._context.topics.subscribe(topic, callback=func_wrapper)
            return func

        return recv_decorator

    def publish_event(self, event: EventSignal):
        return self._context.publish_event(event)

    def trigger_workflow(self, trigger: TriggerSignal):
        return self._context.trigger_workflow(trigger)

    def mutate_state(self, state_patch: TState):
        return self._context.mutate_state(state_patch)

    def channel_send(self, topic: TTopic, data: Any):
        return self._context.channel_send(topic, data)

    def _run_cron_job(self, cron_id: str):
        return self._context.run_cron_job(cron_id)

    def get_state(self) -> TState:
        return self._context.get_state()

    async def _schedule_cron_jobs(self, cron_jobs: Dict[str, CronExpr]):
        scheduler = CronJobScheduler(cron_jobs=cron_jobs)

        async for cron_id in scheduler.schedule():
            self._run_cron_job(cron_id)


def _with_cond(
    cond: WatchedValue[TState, Any],
    runner: SignalRoutineRunner[ReactiveSignal[TState]],
) -> Callable[[ReactiveSignal[TState]], Awaitable[None]]:
    @rename_function(f"{SignalRoutineType.REACTIVE.value}<{runner.name}>")
    async def reactive_routine_wrapper(signal: ReactiveSignal[TState]):
        if cond(signal.old_state) == cond(signal.new_state):
            return

        return await runner(signal)

    return reactive_routine_wrapper



import asyncio
from typing import Annotated, List

from dotenv import find_dotenv, load_dotenv
from langchain_core.messages import AIMessage, AnyMessage, HumanMessage, SystemMessage
from langchain_core.prompts import ChatPromptTemplate
from langchain_openai import ChatOpenAI
from langgraph.graph import add_messages
from pydantic import BaseModel, Field

from livechain.graph.executor import Workflow
from livechain.graph.func import reactive, root, step, subscribe
from livechain.graph.ops import get_state, mutate_state, trigger_workflow
from livechain.graph.types import EventSignal

load_dotenv(find_dotenv())


class AgentState(BaseModel):
    messages: Annotated[List[AnyMessage], add_messages] = Field(default_factory=list)
    has_started: bool = Field(default=False)


class UserChatEvent(EventSignal):
    message: HumanMessage


class RemindUser(BaseModel):
    task: str = Field(..., description="The task to remind the user about")
    remind_after: int = Field(
        ...,
        description="Number of seconds to wait before reminding the user, default to 3 seconds",
    )

    def to_message(self) -> HumanMessage:
        return HumanMessage(content=f"{self.remind_after} seconds has passed, time to remind user about {self.task}.")


class CallReminder(BaseModel):
    reminder: RemindUser | None = Field(default=None, description="The reminder to call")


@step()
async def init_system_prompt(state: AgentState):
    message = SystemMessage(
        content=(
            "You are a helpful assistant to keep track of tasks for a user. "
            "When user want you to remind him of something, you should reply with acknowledgement."
            "Ack should contain the task and the time to remind the user. Another agent will handle the reminder you responded, don't worry about that."
        )
    )

    # op to mutate the state, specify partial updates
    mutate_state(messages=[message], has_started=True)


@step()
async def chat_with_user():
    llm = ChatOpenAI(model="gpt-4o-mini")

    state = get_state(AgentState)
    # await channel_send("assistant", "\nAssistant: ")
    # async for chunk in llm.astream(state.messages):
    #     chunks.append(chunk.content)
    #     await channel_send("assistant", chunk.content)
    response = await llm.ainvoke(state.messages)
    response.pretty_print()

    await mutate_state(messages=AIMessage(content=response.content))


@reactive(AgentState, lambda state: len(state.messages))
async def check_for_reminders(old_state: AgentState, new_state: AgentState):
    messages = new_state.messages

    if len(messages) == 0:
        print("No messages")
        return

    last_message = messages[-1]
    if not isinstance(last_message, AIMessage):
        print("Last message is not an AIMessage")
        return

    llm = ChatOpenAI(model="gpt-4o-mini")
    prompt = ChatPromptTemplate.from_messages(
        [
            (
                "system",
                "Decide whether to remind the user about a task. Say no if not, call RemindUser if yes.",
            ),
            ("ai", "{message}"),
        ]
    )

    chain = prompt | llm.with_structured_output(CallReminder)
    print(f"Last message: {last_message.content}")
    response: CallReminder = await chain.ainvoke({"message": last_message.content})  # type: ignore
    print(f"Response: {response}")

    if response.reminder is None:
        print(f"Response is not a RemindUser: {response}")
        return

    print(f"Reminding user in {response.reminder.remind_after} seconds")
    await asyncio.sleep(response.reminder.remind_after)
    await mutate_state(messages=[response.reminder.to_message()])
    await trigger_workflow()


@subscribe(UserChatEvent)
async def handle_user_chat(event: UserChatEvent):
    await mutate_state(messages=[event.message])
    await trigger_workflow()


@root()
async def entrypoint():
    state = get_state(AgentState)
    if not state.has_started:
        await init_system_prompt(state)

    await chat_with_user()


workflow = Workflow.from_routines(
    root=entrypoint,
    routines=[
        check_for_reminders,
        handle_user_chat,
    ],
)

executor = workflow.compile(AgentState)


async def main():
    @executor.recv("assistant")
    async def handle_assistant_message(message: str):
        print(message)

    executor.start()

    messages = [
        "Set a reminder to buy groceries in 5 seconds",
        5,
        "What is the weather in Tokyo?",
    ]
    # Main event loop that processes input when available
    for message in messages:
        if isinstance(message, int):
            await asyncio.sleep(message)
            continue

        print(f"You: {message}")
        await executor.publish_event(UserChatEvent(message=HumanMessage(content=message)))
        await asyncio.sleep(12)

    await asyncio.sleep(3)
    executor.stop()


if __name__ == "__main__":
    asyncio.run(main())



from asyncio import Task
from typing import Any, Dict, Literal, Optional, Type, overload

from langgraph.config import get_config as get_langgraph_config

from livechain.graph.constants import CONF, CONFIG_KEY_CONTEXT
from livechain.graph.context import Context
from livechain.graph.types import EventSignal, TConfig, TriggerSignal, TState

GraphOp = Literal[
    "get_state",
    "mutate_state",
    "channel_send",
    "publish_event",
    "trigger_workflow",
]


OpResult = Task[Dict[str, Any]]


def get_context(op: GraphOp) -> Context:
    config = get_langgraph_config()

    if CONFIG_KEY_CONTEXT not in config.get(CONF, {}):
        raise RuntimeError(f"Called {op} outside of a workflow")

    return config.get(CONF, {})[CONFIG_KEY_CONTEXT]


def get_config(config_schema: Type[TConfig]) -> TConfig:
    config = get_langgraph_config()

    configurable = config.get(CONF, {})

    return config_schema.model_validate(configurable)


def get_state(state_schema: Type[TState], validate: bool = False) -> TState:
    context = get_context("get_state")
    state = context.get_state()

    if validate:
        return state_schema.model_validate(state)

    return state


@overload
def mutate_state(state_patch: Dict[str, Any]) -> OpResult: ...


@overload
def mutate_state(**kwargs: Any) -> OpResult: ...


def mutate_state(state_patch: Optional[Dict[str, Any]] = None, **kwargs: Any) -> OpResult:
    if state_patch is None:
        return _mutate_state(kwargs)
    else:
        return _mutate_state(state_patch)


def _mutate_state(state_patch: Dict[str, Any]) -> Task[Dict[str, Any]]:
    context = get_context("mutate_state")
    return context.mutate_state(state_patch)


def channel_send(topic: str, data: Any) -> OpResult:
    context = get_context("channel_send")
    return context.channel_send(topic, data)


def publish_event(event: EventSignal) -> OpResult:
    context = get_context("publish_event")
    return context.publish_event(event)


def trigger_workflow() -> OpResult:
    context = get_context("trigger_workflow")
    return context.trigger_workflow(TriggerSignal())



from typing import Any, Dict, Generic, Optional, Type

from pydantic import BaseModel, PrivateAttr

from livechain.graph.emitter import Emitter, emitter_factory
from livechain.graph.persist.base import BaseStatePersister
from livechain.graph.types import CronSignal, EventSignal, ReactiveSignal, TopicSignal, TriggerSignal, TState


def create_default_persister(state_schema: Type[TState]) -> BaseStatePersister[TState]:
    from livechain.graph.persist.local import LocalStatePersister

    return LocalStatePersister(state_schema)


class Context(BaseModel, Generic[TState]):
    state_schema: Type[TState]

    _persister: BaseStatePersister[TState] = PrivateAttr()

    _topic_emitter: Emitter[str, TopicSignal] = PrivateAttr(default_factory=emitter_factory(lambda x: x.topic))

    _event_emitter: Emitter[Type[EventSignal], EventSignal] = PrivateAttr(
        default_factory=emitter_factory(lambda x: type(x))
    )

    _effect_emitter: Emitter[None, ReactiveSignal[TState]] = PrivateAttr(
        default_factory=emitter_factory(lambda _: None)
    )

    _cron_job_emitter: Emitter[str, CronSignal] = PrivateAttr(default_factory=emitter_factory(lambda x: x.cron_id))

    _trigger_emitter: Emitter[None, TriggerSignal] = PrivateAttr(default_factory=emitter_factory(lambda _: None))

    def __init__(
        self,
        state_schema: Type[TState],
        persister: Optional[BaseStatePersister[TState]] = None,
    ):
        super().__init__(state_schema=state_schema)
        self._persister = persister or create_default_persister(state_schema)

    def get_state(self) -> TState:
        return self._persister.get()

    def mutate_state(self, state_patch: TState | Dict[str, Any]):
        prev_state = self._persister.get()
        curr_state = self._persister.set(state_patch)
        state_change = ReactiveSignal(old_state=prev_state, new_state=curr_state)
        return self._effect_emitter.emit(state_change)

    def channel_send(self, topic: str, data: Any):
        return self._topic_emitter.emit(TopicSignal(topic=topic, data=data))

    def publish_event(self, event: EventSignal):
        return self._event_emitter.emit(event)

    def trigger_workflow(self, trigger: TriggerSignal):
        return self._trigger_emitter.emit(trigger)

    def run_cron_job(self, cron_id: str):
        return self._cron_job_emitter.emit(CronSignal(cron_id=cron_id))

    @property
    def events(self):
        return self._event_emitter

    @property
    def effects(self):
        return self._effect_emitter

    @property
    def topics(self):
        return self._topic_emitter

    @property
    def cron_jobs(self):
        return self._cron_job_emitter

    @property
    def trigger(self):
        return self._trigger_emitter



import asyncio
from typing import Annotated, Dict, List, Literal, cast

from dotenv import find_dotenv, load_dotenv
from langchain_core.messages import AnyMessage, HumanMessage, SystemMessage
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_openai import ChatOpenAI
from langgraph.graph import add_messages
from pydantic import BaseModel, Field

from livechain.graph.executor import Workflow, WorkflowExecutor
from livechain.graph.func import root, step, subscribe
from livechain.graph.ops import channel_send, get_config, get_state, mutate_state, trigger_workflow
from livechain.graph.types import EventSignal, TriggerSignal

load_dotenv(find_dotenv())


class AgentState(BaseModel):
    messages: Annotated[List[AnyMessage], add_messages] = Field(default_factory=list)


AgentType = Literal["llm-researcher", "llm-engineer", "gen-ai-engineer"]


class AgentConfig(BaseModel):
    agent_type: AgentType


def agent_type_to_name(agent_type: AgentType) -> str:
    if agent_type == "llm-researcher":
        return "Alex"
    elif agent_type == "llm-engineer":
        return "Bob"
    elif agent_type == "gen-ai-engineer":
        return "Charlie"


def agent_type_to_sys_prompt(agent_type: AgentType) -> str:
    if agent_type == "llm-researcher":
        prompt = (
            "You are a llm researcher, you work at OpenAI, your job involves building and training more advanced transformer models, "
            "such as GPT-5. You are an expert behind math and various training algorithms such as GRPO, RLHF, MoE, etc. "
            "You are invited to participate in a conversation with a llm engineer, and a gen ai product engineer to discuss the future of ai. Asking questions to the other participants to build a better future."
        )
    elif agent_type == "llm-engineer":
        prompt = (
            "You are a llm engineer, you work at Meta, your job involves building and improving LLM serving frameworks and infrastructure, "
            "such as vLLM, SGLang, etc. You are expert in writing different kind of CUDA kernels, quantizations, inference optimizations tricks "
            "such as continuious batching, zero-overhead GPU CPU communication, efficient KV cache aware routings, etc. "
            "You are invited to participate in a conversation with a llm researcher, and a gen ai product engineer to discuss the future of ai. Asking questions to the other participants to build a better future."
        )
    elif agent_type == "gen-ai-engineer":
        prompt = (
            "You are a gen ai product engineer, you work at a Startup, your job involves building and improving gen ai products, "
            "using LLM application frameworks such as LangChain, LangGraph, LlamaIndex, etc. You are an expert in prompt engineering, "
            "RAG, agentic architectures, and LLM quality benchmarkings. You also have a great sense of product design and a deep understanding of user experience. "
            "You are invited to participate in a conversation with a llm researcher, and a llm engineer to discuss the future of ai. Asking questions to the other participants to build a better future."
        )
    else:
        raise ValueError(f"Invalid agent type: {agent_type}")

    prompt += f"Your name is {agent_type_to_name(agent_type)}. At the beginning of the conversation, you will introduce yourself to the other participants. "
    return prompt


class UserChatEvent(EventSignal):
    agent_type: AgentType
    message: str
    should_speak: bool


@step()
async def chat():
    state = get_state(AgentState)
    config = get_config(AgentConfig)

    sys_prompt = agent_type_to_sys_prompt(config.agent_type)
    llm = ChatOpenAI(model="gpt-4o")

    prompt = ChatPromptTemplate.from_messages(
        [
            SystemMessage(content=sys_prompt),
            MessagesPlaceholder(variable_name="messages"),
        ]
    )

    chain = prompt | llm

    response = chain.invoke({"messages": state.messages})
    response.pretty_print()

    await mutate_state({"messages": [response]})
    await channel_send("chat", {"type": config.agent_type, "message": response.content})


@subscribe(UserChatEvent)
async def on_user_chat(event: UserChatEvent):
    name = agent_type_to_name(event.agent_type)
    message = HumanMessage(content=f"{name}: {event.message}")
    await mutate_state({"messages": [message]})

    if event.should_speak:
        await trigger_workflow()


@root()
async def entrypoint():
    chat()


wf = Workflow.from_routines(entrypoint, [on_user_chat])


async def main():
    agent_executors: Dict[AgentType, WorkflowExecutor] = {}

    for agent_type in ["llm-researcher", "llm-engineer", "gen-ai-engineer"]:
        agent_type = cast(AgentType, agent_type)
        executor = wf.compile(AgentState, config_schema=AgentConfig)
        agent_executors[agent_type] = executor

    keys: List[AgentType] = list(agent_executors.keys())
    for agent_type in keys:
        curr_agent_executor = agent_executors[agent_type]

        @curr_agent_executor.recv("chat")
        async def on_chat(data: Dict):
            agent_type = cast(AgentType, data["type"])

            for other_agent_type in keys:
                if other_agent_type == agent_type:
                    continue

                # only next agent in the circle should speak
                should_speak = (keys.index(agent_type) + 1) % len(keys) == keys.index(other_agent_type)

                other_agent_executor = agent_executors[other_agent_type]
                await other_agent_executor.publish_event(
                    UserChatEvent(
                        agent_type=data["type"],
                        message=data["message"],
                        should_speak=should_speak,
                    )
                )

    for agent_type in keys:
        executor = agent_executors[agent_type]
        executor.start(config=AgentConfig(agent_type=agent_type))

    first_agent_executor = agent_executors[keys[0]]
    first_agent_executor.trigger_workflow(TriggerSignal())

    while True:
        await asyncio.sleep(1)


if __name__ == "__main__":
    asyncio.run(main())
